namespace NXyn.Aux
    open System
    
    module IOMisc =
        let StdErr = IO.TextWriter.Synchronized(stderr)

        let Write_Char   (w:IO.TextWriter) (c:Char)   = w.Write(c)
        let Write_Int    (w:IO.TextWriter) (i:Int32)  = w.Write(i)
        let Write_String (w:IO.TextWriter) (s:String) = w.Write(s)
                
        let WriteChar_to    (c:Char)   (w:IO.TextWriter) = w.Write(c)
        let WriteString_to  (s:String) (w:IO.TextWriter) = w.Write(s)
        let WriteNewLine_to            (w:IO.TextWriter) = w.WriteLine("")

        let string_aI l r n = if n < 1 then ""
                                       else let mutable sb = (new Text.StringBuilder()).Append(l:string).Append("a0")
                                            for i = 1 to n - 1 do sb <- sb.Append(',').Append('a').Append(i)
                                            sb.Append(r:string).ToString()

        let WriteMany_to l sep r write1 xs w = if xs <> [||] then Write_String w l
                                                                  write1 xs.[0] w
                                                                  for i = 1 to Array.length xs - 1 do Write_String w sep; write1 xs.[i] w
                                                                  Write_String w r

        let StringMany l sep r to_string xs = if xs = [||] then ""
                                                           else let w = new IO.StringWriter()
                                                                WriteMany_to l sep r (fun el w -> let s = to_string el in w.Write(s:string)) xs w
                                                                w.ToString()


        let WriteManyI_to l sep r write1 xs w = if xs <> [||] then Write_String w l
                                                                   write1 0 xs.[0] w
                                                                   for i = 1 to Array.length xs - 1 do Write_String w sep; write1 i xs.[i] w
                                                                   Write_String w r

        let RemoveFileExtension(s:String) = let p = s.LastIndexOf(".")
                                            if p < 0 then failwith "removeFileExtension"
                                                     else s.Substring(0,p)

        let GetBasedTempFileName (baseDirectory:String) (extension:String) (similar:String) =
            let similar = IO.Path.GetFileNameWithoutExtension(similar)
            let extension = if extension.StartsWith(".") then extension.Substring(1,extension.Length-1) else extension
            let tempPath = sprintf "%s%s/" baseDirectory (IO.Directory.GetCurrentDirectory())
            tempPath |> IO.Directory.CreateDirectory |> ignore

            let rec loop(n:int) = let fullFileName = sprintf "%s%s.%d.%d.%s" tempPath similar ProcessMisc.PId n extension
                                  if IO.File.Exists(fullFileName) then loop(n+1)
                                                                  else let fd = IO.File.Create(fullFileName) in fd.Close(); fullFileName
            loop(1)


        let NXCCbaseDir = IO.Path.GetTempPath() + "nxcc"
        let GetNXCCtmpFileName extension similarTo = GetBasedTempFileName NXCCbaseDir extension similarTo
