namespace NXyn.Aux
    open System
    open System.Collections
    open System.Collections.Generic
    
    module LimitedTextWriter =

        (* TBD: also consider checking for maxrows *)
        type ExnTextWriter(maxcols:int) =
            class
                inherit IO.StringWriter()
                let mutable w = 0
                let mutable h = 0
                let mutable current_width = 0

                let do_check() = if current_width > maxcols then failwith "LimitedTextWriter.add_char: too many columns"
                let newline()  = h <- h + 1; w <- max w current_width; current_width <- 0
                let add_char(c:Char) = if c = '\r' then ()
                                       elif c = '\n' then newline()
                                       else current_width <- current_width + 1; do_check()

                member x.GetWidth()  = max current_width w
                member x.GetHeight() = if current_width = 0 then h else (h + 1)

                override x.ToString() = sprintf "LimitedTextWriter(w=%d,h=%d)" (x.GetWidth()) (x.GetHeight())

                override x.Write(c:Char) = add_char c
                override x.Write(v:String) = for i = 0 to v.Length - 1 do add_char v.[i] done
                override x.WriteLine(c:Char) = x.Write(c); newline ()
                override x.WriteLine(s:String) = x.Write(s); newline ()
            end
