namespace NXyn.EmitFSharp

  open System

  open System.Collections.Generic

  open System.Reflection
  open Microsoft.FSharp.Reflection

  open NXyn.EmitFSharp.Common

  module FoldAttributes =
    let GetUnionCasesAndFields (t:Type) =
        Array.map (fun (uc:UnionCaseInfo) -> let fs = uc.GetFields() in (uc,fs)) (FSharpType.GetUnionCases(t))

    let OutputCode (name:string) allTypes (t:Type) (att_type:Type) (userMapper:obj -> string) (prefix:string) (folderOption:string option) (stateOption:string option) (w:IO.TextWriter) =

      (*XXX*)
      let rec loop newTypes =
        let isIdentity_of_UnionCaseAndFields (t:System.Type) ((uc,fs):(UnionCaseInfo * PropertyInfo [])) =
          Array.forall (fun (f:PropertyInfo) -> f.PropertyType = t || List.forall (fun x -> x <> f.PropertyType) newTypes) fs
          && Array.forall (fun x -> (userMapper x).Split([| ',' |]) = [| |]) (uc.GetCustomAttributes att_type)
        let isIdentity_of_type (t:Type) =
          if   FSharpType.IsRecord t then Array.forall (fun (f:PropertyInfo) -> List.forall ((<>)f.PropertyType) newTypes) (FSharpType.GetRecordFields(t))
          elif FSharpType.IsUnion  t then Array.forall (isIdentity_of_UnionCaseAndFields t) (GetUnionCasesAndFields(t))
          elif FSharpType.IsTuple  t then failwith (Printf.sprintf "EmitFSharpFoldAtts.not_implemented TUPLE %A" t)
          elif FSharpType.IsFunction  t then failwith (Printf.sprintf "EmitFSharpFoldAtts.not_implemented FUNCTION %A" t)
          elif FSharpType.IsModule  t then failwith (Printf.sprintf "EmitFSharpFoldAtts.not_implemented MODULE %A" t)
          elif typeof<bigint> = t then true (* failwith "BIGINT" *)
          elif typeof<string> = t then true (* failwith "BIGINT" *)
          else failwith (Printf.sprintf "EmitFSharpFoldAtts.not_implemented %A" t)
        let newTypes1 = List.filter (fun (t:Type) -> not (isIdentity_of_type t)) newTypes
        if List.length newTypes1 < List.length newTypes then loop newTypes1 else newTypes

      let arg1 = allTypes (* (CollectTypes t) *)
      let res = loop (List.filter (fun (k:Type) -> not (List.exists (fun (t',_) -> t' = k) BuiltinFSharpTypeNames)) arg1)

      match res with
        | [] ->
            match (folderOption,stateOption) with
                | (Some f,Some s) -> fprintfn w "let %s arg = %s"          name s
                | (Some f,None)   -> fprintfn w "let %s s arg = s"         name
                | (None,  Some s) -> fprintfn w "let %s fold_it arg = %s"  name s
                | (None,  None)   -> fprintfn w "let %s fold_it s arg = s" name
        | newTypes ->
            let fold_it = match folderOption with Some s -> s | None -> "fold_it"

            let map = new Dictionary<Type,string>()
            List.iteri (fun (index:int) (t:Type) -> map.Add(t, "fold_" + string index)) newTypes

            let writeFold_of_argument i (f:PropertyInfo) = if map.ContainsKey f.PropertyType then w.Write("let s = {0} s a{1} in ", map.[f.PropertyType], i)

            let writeFold_of_type (t:Type) =
              if FSharpType.IsRecord t then
                let fs = FSharpType.GetRecordFields(t)
                WriteList w (map.[t] + " s ( { ") "; " (" } : " + (FSharpTypeName t) + " ) = ") fs.Length (fun i -> fs.[i].Name + "=a" + string i)

                Array.iteri writeFold_of_argument fs
                w.WriteLine("s")

              elif FSharpType.IsUnion t then
                w.WriteLine(map.[t] + " s = function")

                let some_notcovered = ref false

                for ((uc,fs):(UnionCaseInfo * PropertyInfo [])) in GetUnionCasesAndFields(t) do
                  let ys = ref [] (* Array.concat *)
                  for x in uc.GetCustomAttributes att_type do
                    for y in (userMapper x).Split([| ',' |]) do
                      ys := y :: !ys

                  (* what does "DOES_REC" mean? *)
                  let does_rec = Array.exists (fun (f:PropertyInfo) -> map.ContainsKey f.PropertyType) fs
                  if !ys = [] && (not does_rec) then some_notcovered := true
                  else
                    (* DUP EmitFSharpMap *)
                    let t_name = if t.Name = "FSharpOption`1" then "Option" else t.Name

                    fprintf w "    | %s.%s" t_name uc.Name

                    if not does_rec then w.Write(if fs=[||] then " -> " else " _ -> ") else WriteList w "(" "," ") -> " fs.Length (fun i -> "a" + string i)

                    for y in !ys do
                      fprintf w "let s = (%s) s %s%s in " fold_it prefix y

                    if does_rec then Array.iteri writeFold_of_argument fs
                    fprintfn w "s"

                if !some_notcovered then fprintfn w "    | _ -> s"

              else fprintfn w "%s (arg:%s) = arg" map.[t] (FSharpTypeName t)

            fprintfn w "let %s%s =" name (if Option.isNone(folderOption) then " fold_it" else "")
            List.iteri (fun (i:int) (t:Type) -> w.Write(if i = 0 then "  let rec " else "  and "); writeFold_of_type t) newTypes

            match stateOption with
                | None    -> fprintfn w "  fun s arg -> %s s arg" map.[t]
                | Some(s) -> fprintfn w "  fun arg -> %s %s arg" map.[t] s
      fprintfn w "\n"
