namespace NXyn.EmitFSharp
  
  open System
  open System.Collections.Generic
  
  open System.Reflection
  open Microsoft.FSharp.Reflection
  
  open NXyn.EmitFSharp.Common

  open NXyn.Aux.IOMisc

  (* what about arrays, lists, options, ... *)
  module Map =
    let OutputCode name newTypes t mtype userFilter (w:IO.TextWriter) =
      let map_it = "m"
      match RemoveIdenticalTypes mtype userFilter newTypes with
        | [] -> fprintfn w "let %s %s (x:%s) = x" name map_it (FSharpTypeName(t))
        | ts ->
            let map = new Dictionary<Type,string>()
            ts |> List.iteri (fun i (t:Type) -> map.Add(t, sprintf "m%d" i))

            let writeMap_of_type (t:Type) =
              let tName = fsharpTypeName_of_Type(t) (* ClippedFSharpTypeName(t.Name) *)
              if not(map.ContainsKey(t)) then failwithf "record not contained %s" tName

              if FSharpType.IsUnion t then
                let writeMapIthUCfield i (f:PropertyInfo) w = let t = f.PropertyType
                                                              if   t = mtype && userFilter(f) then fprintf w "%s(a%d)" map_it  i
                                                              elif map.ContainsKey(t)         then fprintf w "%s(a%d)" map.[t] i
                                                                                              else fprintf w "a%d"             i
                fprintfn w "%s(x:%s) =" map.[t] tName
                fprintfn w "    match x with"

                (* HACK *)
                let tName = FSharpTypeName'(t)


                let incomplete = ref false
                FSharpType.GetUnionCases(t)
                |> Array.iter
                     (fun(uc:UnionCaseInfo) ->
                        let fs = uc.GetFields()
                        if Array.exists (fun(f:PropertyInfo) -> map.ContainsKey f.PropertyType (* BOILER *)
                                                                || f.PropertyType = mtype && userFilter f) fs then
                            fprintfn w "      | %s.%s%s -> %s.%s%t" tName uc.Name (string_aI "(" ")" fs.Length)
                                                                    tName uc.Name (WriteManyI_to "(" "," ")" writeMapIthUCfield fs)
                        else incomplete := true)
                if !incomplete then fprintfn w "      | x -> x"

              elif FSharpType.IsRecord t then
                let fs0 = FSharpType.GetRecordFields(t)

                (* filter out which ones stay the same. if there is one, use { x with ... } *)
                let fieldRemainsAsIs(f:PropertyInfo) = let t = f.PropertyType
                                                       if   t = mtype && userFilter(f) then false
                                                       elif map.ContainsKey(t)         then false
                                                                                       else true
                let writeMapRecordField (f:PropertyInfo) w = if   f.PropertyType = mtype && userFilter(f) then fprintf w "%s=%s(x.%s)" f.Name map_it  f.Name
                                                             elif map.ContainsKey(f.PropertyType)         then fprintf w "%s=%s(x.%s)" f.Name map.[f.PropertyType] f.Name
                (* ??? array.choose *)
                let fs = Array.filter (fieldRemainsAsIs >> not) fs0
                fprintfn w "%s(x:%s) : %s = {%s%t}" map.[t] tName tName (if fs.Length <> fs0.Length then " x with" else "") (WriteMany_to " " "; " " " writeMapRecordField fs)
              else fprintf w "%s (arg:%s) = not_implemented" map.[t] tName

            let mtype_name = FSharpTypeName(mtype)
            fprintfn w "let %s (%s : %s -> %s) =" name map_it mtype_name mtype_name
            ts |> List.iteri (fun i (t:Type) -> fprintf w "  %s " (if i = 0 then "let rec" else "and"); writeMap_of_type t)
            fprintfn w "  fun(x:%s) -> %s(x)" (FSharpTypeName t) map.[t]
      fprintfn w "\n"


