
module Disassembler



open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc


open Basics


let iterOverConsecDataDirectives add finish lines = let doLine (d0,ddLines) = function DataDirective(d,_) as ln -> if d0 = d then (d0,(ln::ddLines))
                                                                                                                             else finish(ddLines);(d,[ln])
                                                                                     | line -> finish(ddLines); add(line); ("",[])
                                                    Array.fold doLine ("",[]) lines |> snd |> finish
//let disassemble2 


let disassembleData_in_codeSections(sections) =

    let ddDict = new Dictionary<ILine list, ILine []>()
 
    (* add all sequences of consecutive data directives in text sections to "ddDict" *)
    sections |> Array.choose (fun(name,_,lines) -> if name = ".text" then Some(lines) else None)
             |> Array.iter (iterOverConsecDataDirectives ignore (fun ls -> if ls <> [] then ddDict.[List.rev(ls)] <- [||]))

    if ddDict.Count = 0 then sections
                        else let ddArray = ArrayMisc.array_of_collection(ddDict.Keys)
 
                             let tmp_S       = IOMisc.GetNXCCtmpFileName ".S"       !Global.InFileName
                             let tmp_o       = IOMisc.GetNXCCtmpFileName ".o"       !Global.InFileName
                             let tmp_objdump = IOMisc.GetNXCCtmpFileName ".objdump" !Global.InFileName

                             let w = new IO.StreamWriter(tmp_S)
                             ddArray |> Array.iteri (fun i lines -> fprintfn w ".global .L%dyn\n.L%dyn:" i i; lines |> List.iter (fun l -> WriteILine_to l w; w.WriteLine("")))
                             w.Close()
    
                             if 0 <> ProcessMisc.invoke_with_args "/tmp/nxcc.log" "as"          [tmp_S; "-o"; tmp_o] then failwith "as"
                             if 0 <> ProcessMisc.invoke_with_args "/tmp/nxcc.log" "nxyn_disass" [tmp_o; tmp_objdump] then failwith "objdump"

                             let out_objdump = tmp_objdump |> IO.File.ReadAllLines |> Array.map (fun(s:String) -> s.Trim())

                             (* process 'objdump' output and try to parse instructions that are non-nops *)
                             let mutable ix = 0
                             for i = 0 to ddArray.Length - 1 do
                                 while not(out_objdump.[ix].EndsWith("yn>:")) do ix <- ix + 1
                                 ix <- ix + 1
                                 let lines = new ResizeArray<ILine>()
                                 while ix < out_objdump.Length && not(out_objdump.[ix].EndsWith("yn>:")) do
                                     let s = out_objdump.[ix]
                                     let p = s.IndexOf(":")
                                     if p >= 0 then let asm = s.Substring(p+1,s.Length-p-1).Trim().ToLower()
                                                    if not(asm.StartsWith("nop")) then match TryParseASTline(asm)
                                                                                        with | None -> LogError(sprintf "got unknown instr from objdump (1): %s ('%s')" asm s)
                                                                                             | Some(l) -> match ILine_of_ELine(l)
                                                                                                           with | Some(x) -> lines.Add(x)
                                                                                                                | None -> LogError(sprintf "got unknown instr from objdump (2): %A" l)
                                     ix <- ix + 1
                                 ddDict.[ddArray.[i]] <- lines.ToArray()

                             (* insert new instructions into sections *)
                             let mapTextSection(lines0) = let lines = new ResizeArray<ILine>()
                                                          lines0 |> iterOverConsecDataDirectives lines.Add (fun ls -> if ls <> [] then ddDict.[List.rev(ls)] |> Array.iter lines.Add)
                                                          lines.ToArray()
                             sections |> Array.map (fun(name,args,lines) -> (name,args,lines |> if name <> ".text" then id else mapTextSection))

