module LabelGenerator

let labelChars = [|
                  '0';'1';'2';'3';'4';'5';'6';'7';'8';'9';
                  'A';'B';'C';'D';'E';'F';'G';'H';'I';'J';'K';'L';'M';'N';'O';'P';'Q';'R';'S';'T';'U';'V';'W';'X';'Y';'Z';
                  'a';'b';'c';'d';'e';'f';'g';'h';'i';'j';'k';'l';'m';'n';'o';'p';'q';'r';'s';'t';'u';'v';'w';'x';'y';'z';
                  '_';'_'; (* '.' would be fine too, but I like '_' better *)
                  |]

open System
open System.Collections.Generic

type t(prefix:string,initialMinLen:int) =
    let         rng            = new Random()
    let         set            = new HashSet<string>()

    let         maxLabelLength = 1000
    let         maxTries       = 1000

    let         nLengths       = Array.create (maxLabelLength+1) 0
    let         nTries         = Array.create (maxTries+1)       0
    let mutable minLength      = max 1 initialMinLen

    member x.Register(s) = if String.length s > maxLabelLength then failwithf "LabelGenerator.Register: label '%s' is too long" s
                           if set.Add(s) then nLengths.[s.Length] <- nLengths.[s.Length] + 1

    member x.Unregister(s) = if set.Remove(s) then nLengths.[s.Length] <- nLengths.[s.Length] - 1

    member x.DumpInfo(w) = fprintfn w "[LabelGenerator]  nLengths = %A" nLengths
                           fprintfn w "[LabelGenerator]  nTries   = %A" nTries

    member x.Draw() = let rec adaptMinLength(l) = if prefix.Length + l > maxLabelLength then failwith "LabelGenerator.Draw"
                                                  if 7 * nLengths.[prefix.Length + l] > (1 <<< (6 * l)) then adaptMinLength (l+1)
                                                                                                        else minLength <- l; l
                      let chars = Array.create (adaptMinLength minLength) '9'
                      let rec getNext(n) = if n > maxTries then failwith "LabelGenerator.Draw.getNext"
                                           for i = 0 to minLength - 1 do chars.[i] <- labelChars.[rng.Next() &&& 63]
                                           let candidate = prefix + System.String(chars)
                                           if set.Contains(candidate) then getNext (n+1)
                                                                      else nTries.[n] <- nTries.[n] + 1; candidate
                      let lbl = getNext(1)
                      x.Register(lbl)
                      lbl
